use utf8;
use strict;
use warnings FATAL => 'all';

package Xyzzy::Document::Config;

use Clarity -self;

field templatedir => sub { die "TemplateDir not configured\n" };
field fragmentdir => sub { die "FragmentDir not configured\n" };
field contentdir => sub { die "ContentDir not configured\n" };
field stylesheetdir => sub { die "StylesheetDir not configured\n" };

sub set_templatedir {
	my $dir = $_[0];
	die "Missing argument to TemplateDir\n" unless defined $dir;
	stat $dir;
	warn "TemplateDir $dir doesn't exist\n" unless -e _;
	warn "TemplateDir $dir is not a directory\n" unless -d _;
	$self->templatedir($dir);
}

sub set_fragmentdir {
	my $dir = $_[0];
	die "Missing argument to FragmentDir\n" unless defined $dir;
	stat $dir;
	warn "FragmentDir $dir doesn't exist\n" unless -e _;
	warn "FragmentDir $dir is not a directory\n" unless -d _;
	$self->fragmentdir($dir);
}

sub set_contentdir {
	my $dir = $_[0];
	die "Missing argument to ContentDir\n" unless defined $dir;
	stat $dir;
	warn "ContentDir $dir doesn't exist\n" unless -e _;
	warn "ContentDir $dir is not a directory\n" unless -d _;
	$self->contentdir($dir);
}

sub set_stylesheetdir {
	my $dir = $_[0];
	die "Missing argument to StylesheetDir\n" unless defined $dir;
	stat $dir;
	warn "StylesheetDir $dir doesn't exist\n" unless -e _;
	warn "StylesheetDir $dir is not a directory\n" unless -d _;
	$self->stylesheetdir($dir);
}

package Xyzzy::Document;

use XML::LibXML;
use XML::LibXML::XPathContext;
use XML::LibXSLT;

use Clarity -self;
use Xyzzy::Hook -mixin;

field xml => sub { new XML::LibXML::Document("1.0", "UTF-8") };
field cfg => sub { shift->req->cfg };
field req;

sub _gethash() {
	return $_[0]->{$_[1]}
}

sub getElementById {
	my $id = shift;

#	our $expr ||= new XML::LibXML::XPathExpression('//*[@id=$id]');
	my $expr = '//*[@id=$id]';

	our $ctx ||= new XML::LibXML::XPathContext;

	our $prev;
	$ctx->registerVarLookupFunc(\&_gethash, {id => $id})
		unless defined $prev && $prev eq $id;

	$prev = $id;

	my $nodes = $ctx->find($expr, $self->xml);

	confess "Document ID '$id' not unique"
		if @$nodes > 1;

	return $nodes->[0]
}

sub xmlParser {
	our $xmlparser ||= new XML::LibXML;
	$xmlparser->validation(0);
	$xmlparser->load_ext_dtd(0);
	$xmlparser->expand_entities(0);
	$xmlparser->pedantic_parser(1);
	return $xmlparser;
}

sub xsltParser {
	return our $xsltparser ||= new XML::LibXSLT;
}

sub getStylesheet {
	my $file = shift;
	our %stylesheetcache;

	return $stylesheetcache{$file}
		if exists $stylesheetcache{$file};

	my $xmlparser = $self->xmlParser;
	my $xml = $xmlparser->parse_file($file)
		or die "can't parse $file\n";
	my $xsltparser = $self->xsltParser;
	my $xslt = $xsltparser->parse_stylesheet($xml)
		or die "can't parse $file as XSLT\n";

	return $stylesheetcache{$file} = $xslt;
}

sub applyStylesheets {
	my $dir = $self->cfg->stylesheetdir;
	my $xml = $self->xml;
	foreach my $xslt (map { $self->getStylesheet("$dir/$_.xml") } @_) {
		$xml = $xslt->transform($xml)
			or die "XSLT transform failed\n";
	}
	$self->xml($xml);
}

sub construct {
	my $name = shift;
	my @children = @_;
	my $node = $self->createElement($name);
	foreach my $child (@children) {
		if(ref $child) {
			my $type = reftype $child;
			if($type eq 'ARRAY') {
				$node->appendChild($self->construct(@$child));
			} elsif($type eq 'HASH') {
				while(my @attr = each(%$child)) {
					$node->setAttribute(@attr);
				}
			} else {
				$node->appendChild($child);
			}
		} else {
			$node->appendText($child);
		}
	}
	return $node;
}

sub serialize {
	return new Xyzzy::Response(mimetype => 'application/xml', content => $self->toString);
}

sub finalize {
#	$self->xml->validate;
	my $res = $self->serialize;
	delete $self->{xml};
	return $res;
}

sub response {
	$self->hook_call('build');
	return $self->finalize;
}

sub bakecookie {
	my $name = shift;
	my $value = shift;
	my $req = $self->req;
	my %defaults = ( 
			Path => $req->url(-absolute => 1),
			Secure => $req->secure,
		);

	return new Xyzzy::Cookie($name => $value, %defaults, @_);
}

sub DESTROY() {} # don't try to autoload this

sub AUTOLOAD {
	my $off = rindex(our $AUTOLOAD, '::');
	confess "no package name in '$AUTOLOAD'"
		if $off == -1;
	my $sub = substr($AUTOLOAD, $off + 2);
	my $xml = $self->xml;
	my $code = "sub $sub { my \$self = shift; return \$self->xml->$sub(\@_) }";
	my $err = do { local $@; eval $code; $@ };
	confess $err if $err;
	return $xml->$sub(@_);
}
